/*
 *  MICO --- a free CORBA implementation
 *  Copyright (C) 1997-98 Kay Roemer & Arno Puder
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#define MICO_CONF_IMR
#define MICO_CONF_DYNANY

#include <CORBA-SMALL.h>
#include <mico/dynany_impl.h>
#include <mico/impl.h>
#include <mico/throw.h>
#include <mico/template_impl.h>


/************************** DynAny *********************************/


DynAny_impl::DynAny_impl ()
    : _index (0)
{
}

DynAny_impl::~DynAny_impl ()
{
}

CORBA::ORB_ptr
DynAny_impl::_factory ()
{
    static CORBA::ORB_ptr factory =
	CORBA::ORB::_nil();

    if (CORBA::is_nil (factory))
	factory = CORBA::ORB_instance ("mico-local-orb");

    return factory;
}

void
DynAny_impl::update_element (CORBA::Long idx)
{
}

void
DynAny_impl::assign (CORBA::DynAny_ptr dyn_any)
{
    CORBA::Any_var a = dyn_any->to_any ();
    from_any (a);
}

CORBA::DynAny_ptr
DynAny_impl::copy ()
{
    CORBA::Any_var a = to_any ();
    return _factory()->create_dyn_any (a);
}

CORBA::TypeCode_ptr
DynAny_impl::type ()
{
    return CORBA::TypeCode::_duplicate (_type);
}

void
DynAny_impl::destroy ()
{
    for (CORBA::ULong i = 0; i < _elements.size(); ++i)
	_elements[i]->destroy();
    _elements.erase (_elements.begin(), _elements.end());
    CORBA::release (this);
}

/*
 * do insert_* and get_* advance _index? This isn't explicitely
 * stated in the docs, but the given example suggests that at least
 * the insert_* ops do so.
 */

void
DynAny_impl::insert_boolean (CORBA::Boolean value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= CORBA::Any::from_boolean (value);
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_octet (CORBA::Octet value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= CORBA::Any::from_octet (value);
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_char (CORBA::Char value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= CORBA::Any::from_char (value);
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_wchar (CORBA::WChar value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= CORBA::Any::from_wchar (value);
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_short (CORBA::Short value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_ushort (CORBA::UShort value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_long (CORBA::Long value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_longlong (CORBA::LongLong value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_ulong (CORBA::ULong value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_ulonglong (CORBA::ULongLong value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_float (CORBA::Float value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_double (CORBA::Double value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_longdouble (CORBA::LongDouble value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_string (const char* value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_wstring (const CORBA::WChar* value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_reference (CORBA::Object_ptr value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    if (_type->unalias()->kind() == CORBA::tk_objref)
        a <<= CORBA::Any::from_object (value, _type->name());
    else
        a <<= CORBA::Any::from_object (value, "");
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_typecode (CORBA::TypeCode_ptr value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_valuetype (CORBA::ValueBase *value)
{
    update_element (_index);

    CORBA::StaticAny sa (CORBA::_stc_ValueBase, &value);

    CORBA::Any a;
    // take care for aliases
    CORBA::TypeCode_var tc = _elements[_index]->type();

    if (!a.from_static_any (sa, tc))
	mico_throw (Invalid());

    _elements[_index]->from_any (a);
    next ();
}

void
DynAny_impl::insert_any (const CORBA::Any &value)
{
    update_element (_index);
    CORBA::Any a;
    // take care for aliases of base types
    CORBA::TypeCode_var tc = _elements[_index]->type();
    a.type (tc);
    a <<= value;
    _elements[_index]->from_any (a);
    next ();
}

CORBA::Boolean
DynAny_impl::get_boolean ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Boolean value;
    if (!((CORBA::Any &)a >>= CORBA::Any::to_boolean (value)))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Octet
DynAny_impl::get_octet ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Octet value;
    if (!((CORBA::Any &)a >>= CORBA::Any::to_octet (value)))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Char
DynAny_impl::get_char ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Char value;
    if (!((CORBA::Any &)a >>= CORBA::Any::to_char (value)))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::WChar
DynAny_impl::get_wchar ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::WChar value;
    if (!((CORBA::Any &)a >>= CORBA::Any::to_wchar (value)))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Short
DynAny_impl::get_short ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Short value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::UShort
DynAny_impl::get_ushort ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::UShort value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Long
DynAny_impl::get_long ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Long value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::LongLong
DynAny_impl::get_longlong ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::LongLong value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::ULong
DynAny_impl::get_ulong ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::ULong value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::ULongLong
DynAny_impl::get_ulonglong ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::ULongLong value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Float
DynAny_impl::get_float ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Float value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::Double
DynAny_impl::get_double ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Double value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::LongDouble
DynAny_impl::get_longdouble ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::LongDouble value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

char*
DynAny_impl::get_string ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    char *value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return CORBA::string_dup (value);
}

CORBA::WChar*
DynAny_impl::get_wstring ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::WChar *value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return CORBA::wstring_dup (value);
}

CORBA::Object_ptr
DynAny_impl::get_reference ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Object_ptr value;
    if (!((CORBA::Any &)a >>= CORBA::Any::to_object (value)))
	mico_throw (TypeMismatch ());
    next ();
    return CORBA::Object::_duplicate (value);
}

CORBA::TypeCode_ptr
DynAny_impl::get_typecode ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::TypeCode_ptr value;
    if (!((CORBA::Any &)a >>= value))
	mico_throw (TypeMismatch ());
    next ();
    return CORBA::TypeCode::_duplicate (value);
}

CORBA::Any*
DynAny_impl::get_any ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();
    CORBA::Any *value = new CORBA::Any;
    if (!((CORBA::Any &)a >>= *value))
	mico_throw (TypeMismatch ());
    next ();
    return value;
}

CORBA::ValueBase *
DynAny_impl::get_valuetype ()
{
    update_element (_index);
    CORBA::Any_var a = _elements[_index]->to_any();

    CORBA::ValueBase *value;
    CORBA::StaticAny sa (CORBA::_stc_ValueBase, &value);

    if (!a->to_static_any (sa))
	mico_throw (TypeMismatch());

    next ();
    return value;
}

CORBA::DynAny_ptr
DynAny_impl::current_component ()
{
    update_element (_index);
    return CORBA::DynAny::_duplicate (_elements[_index]);
}

CORBA::Boolean
DynAny_impl::next ()
{
    if ((CORBA::ULong)_index+1 == _elements.size())
	return FALSE;
    ++_index;
    return TRUE;
}

CORBA::Boolean
DynAny_impl::seek (CORBA::Long index)
{
    update_element (index); // might cause element to disappear
    if (index < 0 || (CORBA::ULong)index >= _elements.size())
	return FALSE;
    _index = index;
    return TRUE;
}

void
DynAny_impl::rewind ()
{
    _index = 0;
}


/************************* DynBasic *******************************/


DynBasic_impl::DynBasic_impl (CORBA::TypeCode_ptr tc)
{
    _type = CORBA::TypeCode::_duplicate (tc);
    // take care for aliases of base types
    _value.type (_type);
    tc = tc->unalias ();
    switch (tc->kind()) {
    case CORBA::tk_void:
    case CORBA::tk_null:
	break;
    case CORBA::tk_char:
	_value <<= CORBA::Any::from_char (0);
	break;
    case CORBA::tk_wchar:
	_value <<= CORBA::Any::from_wchar (0);
	break;
    case CORBA::tk_boolean:
	_value <<= CORBA::Any::from_boolean (FALSE);
	break;
    case CORBA::tk_octet:
	_value <<= CORBA::Any::from_octet (0);
	break;
    case CORBA::tk_short:
	_value <<= (CORBA::Short)0;
	break;
    case CORBA::tk_ushort:
	_value <<= (CORBA::UShort)0;
	break;
    case CORBA::tk_long:
	_value <<= (CORBA::Long)0;
	break;
    case CORBA::tk_longlong:
	_value <<= (CORBA::LongLong)0;
	break;
    case CORBA::tk_ulong:
	_value <<= (CORBA::ULong)0;
	break;
    case CORBA::tk_ulonglong:
	_value <<= (CORBA::ULongLong)0;
	break;
    case CORBA::tk_float:
	_value <<= (CORBA::Float)0;
	break;
    case CORBA::tk_double:
	_value <<= (CORBA::Double)0;
	break;
    case CORBA::tk_longdouble:
	_value <<= (CORBA::LongDouble)0;
	break;
    case CORBA::tk_any:
	_value <<= CORBA::Any();
	break;
    case CORBA::tk_TypeCode:
	_value <<= CORBA::_tc_void;
	break;
    case CORBA::tk_objref:
	_value <<= CORBA::Any::from_object (CORBA::Object::_nil(), "");
	break;
    case CORBA::tk_string:
        _value <<= "";
        break;
    case CORBA::tk_wstring:
        _value <<= L"";
        break;
    default:
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    }
    _elements.push_back (_duplicate (this));
}

DynBasic_impl::DynBasic_impl (const CORBA::Any &a)
{
    _value = a;
    _type = _value.type();

    switch (_type->unalias()->kind()) {
    case CORBA::tk_void:
    case CORBA::tk_null:
    case CORBA::tk_char:
    case CORBA::tk_wchar:
    case CORBA::tk_boolean:
    case CORBA::tk_octet:
    case CORBA::tk_short:
    case CORBA::tk_ushort:
    case CORBA::tk_long:
    case CORBA::tk_longlong:
    case CORBA::tk_ulong:
    case CORBA::tk_ulonglong:
    case CORBA::tk_float:
    case CORBA::tk_double:
    case CORBA::tk_longdouble:
    case CORBA::tk_any:
    case CORBA::tk_TypeCode:
    case CORBA::tk_objref:
    case CORBA::tk_string:
    case CORBA::tk_wstring:
	break;
    default:
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    }
    _elements.push_back (_duplicate (this));
}

DynBasic_impl::~DynBasic_impl ()
{
}

void
DynBasic_impl::destroy ()
{
    _elements.pop_back ();
    CORBA::release (this);
}

void
DynBasic_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (_type->unalias()->kind() != CORBA::tk_objref ||
	tc->unalias()->kind() != CORBA::tk_objref) {
        if (!_type->equaltype (tc))
            mico_throw (Invalid ());
    }
    _value = value;
}

CORBA::Any*
DynBasic_impl::to_any ()
{
    return new CORBA::Any (_value);
}


/************************** DynFixed ********************************/


DynFixed_impl::DynFixed_impl (CORBA::TypeCode_ptr tc)
{
    if (tc->unalias()->kind() != CORBA::tk_fixed)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    _type = CORBA::TypeCode::_duplicate (tc);

    CORBA::UShort digits = _type->unalias()->fixed_digits();
    CORBA::Short scale = _type->unalias()->fixed_scale();

    Fixed f (digits, scale);
    
    _value <<= CORBA::Any::from_fixed (f, digits, scale);

    _elements.push_back (_duplicate (this));
}

DynFixed_impl::DynFixed_impl (const CORBA::Any &a)
{
    _value = a;
    _type = a.type();
    if (_type->unalias()->kind() != CORBA::tk_fixed)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    _elements.push_back (_duplicate (this));
}

DynFixed_impl::~DynFixed_impl ()
{
}

void
DynFixed_impl::destroy ()
{
    _elements.pop_back ();
    CORBA::release (this);
}

void
DynFixed_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());
    _value = value;
}

CORBA::Any*
DynFixed_impl::to_any ()
{
    return new CORBA::Any (_value);
}

CORBA::OctetSeq*
DynFixed_impl::get_value ()
{
    CORBA::UShort digits = _type->unalias()->fixed_digits();
    CORBA::Short scale = _type->unalias()->fixed_scale();

    Fixed f (digits, scale);

    CORBA::Boolean r = (_value >>= CORBA::Any::to_fixed (f, digits, scale));
    assert (r);

    // convert to CDR format
    MICO::CDREncoder ec;
    FixedBase::FixedValue_var v = f.to_digits();
    ec.put_fixed (v, digits, scale);
    CORBA::OctetSeq *out =
	new CORBA::OctetSeq (ec.buffer()->length(),
			     ec.buffer()->length(),
			     (CORBA::Octet *)ec.buffer()->buffer(),
			     FALSE);
    return out;
}

void
DynFixed_impl::set_value (const CORBA::OctetSeq& in)
{
    CORBA::UShort digits = _type->unalias()->fixed_digits();
    CORBA::Short scale = _type->unalias()->fixed_scale();

    // convert from CDR format
    FixedBase::FixedValue v;
    MICO::CDRDecoder dc;
    dc.buffer()->put (&in[0], in.length());
    // XXX what exception
    CORBA::Boolean r = dc.get_fixed (v, digits, scale);
    assert (r);

    Fixed f (digits, scale);
    f.from_digits (v);

    // XXX what exception
    _value <<= CORBA::Any::from_fixed (f, digits, scale);
}


/************************** DynEnum ********************************/


DynEnum_impl::DynEnum_impl (CORBA::TypeCode_ptr tc)
{
    if (tc->unalias()->kind() != CORBA::tk_enum)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    _type = CORBA::TypeCode::_duplicate (tc);
    _value.type (tc);
    _value.enum_put (0);

    _elements.push_back (_duplicate (this));
}

DynEnum_impl::DynEnum_impl (const CORBA::Any &a)
{
    _value = a;
    _type = a.type();
    if (_type->unalias()->kind() != CORBA::tk_enum)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    _elements.push_back (_duplicate (this));
}

DynEnum_impl::~DynEnum_impl ()
{
}

void
DynEnum_impl::destroy ()
{
    _elements.pop_back ();
    CORBA::release (this);
}

void
DynEnum_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());
    _value = value;
}

CORBA::Any*
DynEnum_impl::to_any ()
{
    return new CORBA::Any (_value);
}

char*
DynEnum_impl::value_as_string ()
{
    return CORBA::string_dup (
	_type->unalias()->member_name (value_as_ulong()));
}

void
DynEnum_impl::value_as_string (const char* value)
{
    CORBA::TypeCode_ptr tc = _type->unalias();
    CORBA::Long idx = tc->member_index (value);
    // XXX what exception ???
    assert (idx >= 0);
    CORBA::Boolean r = _value.enum_put (idx);
    assert (r);
}

CORBA::ULong
DynEnum_impl::value_as_ulong ()
{
    CORBA::ULong l;
    CORBA::Boolean r = _value.enum_get (l);
    assert (r);
    return l;
}

void
DynEnum_impl::value_as_ulong (CORBA::ULong value)
{
    // XXX what exception ???
    CORBA::Boolean r = _value.enum_put (value);
    assert (r);
}


/************************* DynStruct *******************************/


DynStruct_impl::DynStruct_impl (CORBA::TypeCode_ptr tc)
{
    _type = CORBA::TypeCode::_duplicate (tc);
    tc = tc->unalias();
    if (tc->kind() != CORBA::tk_struct && tc->kind() != CORBA::tk_except)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    _isexcept = (tc->kind() == CORBA::tk_except);
    for (CORBA::ULong i = 0; i < tc->member_count(); ++i) {
	CORBA::TypeCode_var memtc = tc->member_type(i);
	_elements.push_back (_factory()->create_dyn_tc (memtc));
    }
}

DynStruct_impl::DynStruct_impl (const CORBA::Any &a)
{
    _type = a.type();
    CORBA::TypeCode_ptr tc = _type->unalias();
    if (tc->kind() != CORBA::tk_struct && tc->kind() != CORBA::tk_except)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    _isexcept = (tc->kind() == CORBA::tk_except);

    if (_isexcept) {
	CORBA::String_var repoid;
	CORBA::Boolean r = a.except_get_begin (repoid.out());
	assert (r);
    } else {
	CORBA::Boolean r = a.struct_get_begin ();
	assert (r);
    }
    for (CORBA::ULong i = 0; i < tc->member_count(); ++i) {
	CORBA::Any el;
	CORBA::Boolean r = a.any_get (el);
	assert (r);
	_elements.push_back (_factory()->create_dyn_any (el));
    }
    if (_isexcept) {
	CORBA::Boolean r = a.except_get_end ();
	assert (r);
    } else {
	CORBA::Boolean r = a.struct_get_end ();
	assert (r);
    }
}

DynStruct_impl::~DynStruct_impl ()
{
}

void
DynStruct_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());

    if (_isexcept) {
	CORBA::String_var repoid;
	CORBA::Boolean r = value.except_get_begin (repoid.out());
	assert (r);
    } else {
	CORBA::Boolean r = value.struct_get_begin ();
	assert (r);
    }
    CORBA::TypeCode_ptr tc2 = tc->unalias();
    for (CORBA::ULong i = 0; i < tc2->member_count(); ++i) {
	CORBA::Any el;
	CORBA::Boolean r = value.any_get (el);
	assert (r);
	_elements[i]->from_any (el);
    }
    if (_isexcept) {
	CORBA::Boolean r = value.except_get_end ();
	assert (r);
    } else {
	CORBA::Boolean r = value.struct_get_end ();
	assert (r);
    }
}

CORBA::Any*
DynStruct_impl::to_any ()
{
    CORBA::Any *a = new CORBA::Any;

    a->type (_type);

    if (_isexcept) {
	CORBA::TypeCode_ptr tc = _type->unalias();
	CORBA::Boolean r = a->except_put_begin (tc->id());
	assert (r);
    } else {
	CORBA::Boolean r = a->struct_put_begin ();
	assert (r);
    }
    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	CORBA::Any_var el = _elements[i]->to_any ();
	CORBA::Boolean r = a->any_put (el);
	assert (r);
    }
    if (_isexcept) {
	CORBA::Boolean r = a->except_put_end ();
	assert (r);
    } else {
	CORBA::Boolean r = a->struct_put_end ();
	assert (r);
    }
    return a;
}

CORBA::FieldName
DynStruct_impl::current_member_name ()
{
    return CORBA::string_dup (_type->unalias()->member_name (_index));
}

CORBA::TCKind
DynStruct_impl::current_member_kind ()
{
    // XXX unlias ???
    CORBA::TypeCode_var tc = _type->unalias()->member_type (_index);
    return tc->kind();
}

CORBA::NameValuePairSeq*
DynStruct_impl::get_members ()
{
    CORBA::NameValuePairSeq *seq = new CORBA::NameValuePairSeq;
    seq->length (_elements.size());
    CORBA::TypeCode_ptr tc = _type->unalias ();
    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	(*seq)[i].id = tc->member_name (i);
	CORBA::Any_var a = _elements[i]->to_any ();
	(*seq)[i].value = a;
    }
    return seq;
}

void
DynStruct_impl::set_members (const CORBA::NameValuePairSeq& value)
{
    CORBA::TypeCode_ptr tc = _type->unalias ();
    if (value.length() != tc->member_count())
	mico_throw (InvalidSeq ());

    for (CORBA::ULong i = 0; i < value.length(); ++i) {
	if (strcmp (tc->member_name(i), value[i].id))
	    mico_throw (InvalidSeq ());
	// XXX throws Invalid() instead of InvalidSeq() ...
	_elements[i]->from_any (value[i].value);
    }
}


/************************* DynUnion *******************************/


DynUnion_impl::DynUnion_impl (CORBA::TypeCode_ptr tc)
{
    _type = CORBA::TypeCode::_duplicate (tc);
    tc = tc->unalias ();
    if (tc->kind() != CORBA::tk_union)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    CORBA::TypeCode_var disctc = tc->discriminator_type();
    _elements.push_back (_factory()->create_dyn_tc (disctc));
    CORBA::Any_var disc = _elements[0]->to_any ();
    _member_idx = tc->member_index (disc);
    if (_member_idx >= 0) {
	CORBA::TypeCode_var memtc = tc->member_type (_member_idx);
	_elements.push_back (_factory()->create_dyn_tc (memtc));
    } else {
	// XXX what to do for implicit default case ???
    }
}

DynUnion_impl::DynUnion_impl (const CORBA::Any &a)
{
    _type = a.type();
    CORBA::TypeCode_ptr tc = _type->unalias ();
    if (tc->kind() != CORBA::tk_union)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    CORBA::Boolean r = a.union_get_begin ();
    assert (r);
    CORBA::Any disc;
    r = a.any_get (disc);
    assert (r);
    _elements.push_back (_factory()->create_dyn_any (disc));
    _member_idx = tc->member_index (disc);
    if (_member_idx >= 0) {
	r = a.union_get_selection (_member_idx);
	assert (r);
	CORBA::Any el;
	r = a.any_get (el);
	assert (r);
	_elements.push_back (_factory()->create_dyn_any (el));
    } else {
	// XXX what to do for implicit default case ???
    }
    r = a.union_get_end ();
    assert (r);
}

DynUnion_impl::~DynUnion_impl ()
{
}

void
DynUnion_impl::update_element (CORBA::Long idx)
{
    if (idx != 1)
	return;

    // XXX slow 
    CORBA::Any_var disc = _elements[0]->to_any ();
    CORBA::Long nidx = _type->unalias()->member_index (disc);
    if (nidx == _member_idx)
	return;

    if (_elements.size() == 2)
	_elements.pop_back ();
    if (nidx >= 0) {
	CORBA::TypeCode_var tc = _type->unalias()->member_type (nidx);
	_elements.push_back (_factory()->create_dyn_tc (tc));
    } else {
	// XXX what to do for implicit default case ???
    }
    _member_idx = nidx;

    if (_index >= (CORBA::Long)_elements.size())
	_index = _elements.size()-1;
}

void
DynUnion_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());

    CORBA::Boolean r = value.union_get_begin ();
    assert (r);
    CORBA::Any disc;
    r = value.any_get (disc);
    assert (r);
    _elements[0]->from_any (disc);
    update_element (1);
    if (_member_idx >= 0) {
	r = value.union_get_selection (_member_idx);
	assert (r);
	CORBA::Any el;
	r = value.any_get (el);
	assert (r);
	_elements[1]->from_any (el);
    } else {
	// XXX what to do for implicit default case ???
    }
    r = value.union_get_end ();
    assert (r);
}

CORBA::Any*
DynUnion_impl::to_any ()
{
    if (set_as_default()) {
      set_as_default (TRUE);
    }

    CORBA::Any *a = new CORBA::Any;

    a->type (_type);
    CORBA::Boolean r = a->union_put_begin ();
    assert (r);
    CORBA::Any_var disc = _elements[0]->to_any ();
    r = a->any_put (disc);
    assert (r);
    CORBA::Long idx = _type->unalias()->member_index (disc);
    update_element (1);
    if (idx >= 0) {
	r = a->union_put_selection (idx);
	assert (r);
	CORBA::Any_var val = _elements[1]->to_any ();
	r = a->any_put (val);
	assert (r);
    }
    r = a->union_put_end ();
    assert (r);
    return a;
}

CORBA::Boolean
DynUnion_impl::set_as_default ()
{
    CORBA::Any_var disc = _elements[0]->to_any ();
    CORBA::Long idx = _type->unalias()->member_index (disc);
    return idx < 0 || idx == _type->unalias()->default_index();
}

void
DynUnion_impl::set_as_default (CORBA::Boolean value)
{
    CORBA::TypeCode_ptr tc = _type->unalias();
    CORBA::Long defidx = tc->default_index();

    if (value) {
	CORBA::TypeCode_var disc = tc->discriminator_type();
	disc = CORBA::TypeCode::_duplicate (disc->unalias());

	if (disc->kind() == CORBA::tk_enum) {
            CORBA::Any any;
            any.type (disc);
	    for (CORBA::ULong i = 0; i < disc->member_count(); ++i) {
		CORBA::Boolean r = any.enum_put (i);
		assert (r);
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    }
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_long) {
	    CORBA::Long l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_longlong) {
	    CORBA::LongLong l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_ulong) {
	    CORBA::ULong l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_ulonglong) {
	    CORBA::ULongLong l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_char) {
	    CORBA::Char l = 0;
            CORBA::Any any;
	    do {
		any <<= CORBA::Any::from_char (l);
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_wchar) {
	    CORBA::WChar l = 0;
            CORBA::Any any;
	    do {
		any <<= CORBA::Any::from_wchar (l);
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_short) {
	    CORBA::Short l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_ushort) {
#if !defined(__GNUG__) && defined(__sgi)
	    volatile
#endif
	    CORBA::UShort l = 0;
            CORBA::Any any;
	    do {
		any <<= l;
		if (tc->member_index (any) == defidx) {
		    _elements[0]->from_any (any);
		    return;
		}
	    } while (++l);
	    // XXX what exception ???
	    assert (0);
	}
	if (disc->kind() == CORBA::tk_boolean) {
	    CORBA::Any any;
	    any <<= CORBA::Any::from_boolean (TRUE);
	    if (tc->member_index (any) == defidx) {
		_elements[0]->from_any (any);
		return;
	    }
	    any <<= CORBA::Any::from_boolean (FALSE);
	    if (tc->member_index (any) == defidx) {
		_elements[0]->from_any (any);
		return;
	    }
	    // XXX what exception ???
	    assert (0);
	}
    } else {
	for (CORBA::ULong i = 0; i < tc->member_count(); ++i) {
	    if (i != (CORBA::ULong)defidx) {
		CORBA::Any_var mem = tc->member_label (i);
		_elements[0]->from_any (mem);
		return;
	    }
	}
	// unreachable, since empty union not allowed
	assert (0);
    }
}

CORBA::FieldName
DynUnion_impl::member_name ()
{
    CORBA::Any_var disc = _elements[0]->to_any ();
    CORBA::Long idx = _type->unalias()->member_index (disc);
    if (idx < 0)
	return CORBA::string_dup ("");
    return CORBA::string_dup (_type->unalias()->member_name (idx));
}

void
DynUnion_impl::member_name (const char *value)
{
    CORBA::TypeCode_ptr tc = _type->unalias();
    CORBA::Long idx = tc->member_index (value);
    assert (idx >= 0);
    if (idx == tc->default_index()) {
	set_as_default (TRUE);
    } else {
	CORBA::Any_var mem = tc->member_label(idx);
	_elements[0]->from_any (mem);
    }
}

CORBA::DynAny_ptr
DynUnion_impl::discriminator ()
{
    return CORBA::DynAny::_duplicate (_elements[0]);
}

CORBA::TCKind
DynUnion_impl::discriminator_kind ()
{
    // XXX unalias ???
    CORBA::TypeCode_var tc = _elements[0]->type();
    return tc->kind();
}

CORBA::DynAny_ptr
DynUnion_impl::member ()
{
    // XXX implicit default has no value ???
    update_element (1);
    return CORBA::DynAny::_duplicate (_elements[1]);
}

CORBA::TCKind
DynUnion_impl::member_kind ()
{
    // XXX unalias ???
    // XXX implicit default has no value ???
    update_element (1);
    CORBA::TypeCode_var tc = _elements[1]->type();
    return tc->kind();
}


/************************ DynSequence *****************************/


DynSequence_impl::DynSequence_impl (CORBA::TypeCode_ptr tc)
{
    _type = CORBA::TypeCode::_duplicate (tc);
    tc = tc->unalias ();
    if (tc->kind() != CORBA::tk_sequence)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    _length = 0;
}

DynSequence_impl::DynSequence_impl (const CORBA::Any &a)
{
    _type = a.type ();
    CORBA::TypeCode_ptr tc = _type->unalias ();
    if (tc->kind() != CORBA::tk_sequence)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    CORBA::Boolean r = a.seq_get_begin (_length);
    assert (r);
    for (CORBA::ULong i = 0; i < _length; ++i) {
	CORBA::Any el;
	r = a.any_get (el);
	assert (r);
	_elements.push_back (_factory()->create_dyn_any (el));
    }
    r = a.seq_get_end ();
    assert (r);
}

DynSequence_impl::~DynSequence_impl ()
{
}

void
DynSequence_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());

    CORBA::ULong len;
    CORBA::Boolean r = value.seq_get_begin (len);
    assert (r);
    if (len != _length)
	length (len);
    for (CORBA::ULong i = 0; i < _length; ++i) {
	CORBA::Any el;
	r = value.any_get (el);
	assert (r);
	_elements[i]->from_any (el);
    }
    r = value.seq_get_end ();
    assert (r);
}

CORBA::Any*
DynSequence_impl::to_any ()
{
    CORBA::Any *a = new CORBA::Any;

    a->type (_type);
    CORBA::Boolean r = a->seq_put_begin (_length);
    assert (r);
    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	CORBA::Any_var el = _elements[i]->to_any();
	r = a->any_put (el);
	assert (r);
    }
    r = a->seq_put_end ();
    assert (r);
    return a;
}

CORBA::ULong
DynSequence_impl::length ()
{
    return _length;
}

void
DynSequence_impl::length (CORBA::ULong value)
{
    if (value < _elements.size ()) {
	_elements.erase (_elements.begin() + value, _elements.end());
	if (value == 0)
	    _index = 0;
	else if (_index >= (CORBA::Long)value)
	    _index = value-1;
    } else if (value > _elements.size()) {
	CORBA::TypeCode_var tc = _type->unalias()->content_type ();
	for (CORBA::ULong i = 0; i < value - _length; ++i)
	    _elements.push_back (_factory()->create_dyn_tc (tc));
    }
    _length = value;
}

CORBA::AnySeq*
DynSequence_impl::get_elements ()
{
    CORBA::AnySeq *seq = new CORBA::AnySeq;
    seq->length (_length);

    for (CORBA::ULong i = 0; i < _length; ++i) {
	CORBA::Any_var el = _elements[i]->to_any ();
	(*seq)[i] = el;
    }
    return seq;
}

void
DynSequence_impl::set_elements (const CORBA::AnySeq& value)
{
    if (value.length() != _length)
	mico_throw (InvalidSeq ());

    for (CORBA::ULong i = 0; i < _length; ++i) {
	_elements[i]->from_any (value[i]);
    }
}


/************************** DynArray *******************************/


DynArray_impl::DynArray_impl (CORBA::TypeCode_ptr tc)
{
    _type = CORBA::TypeCode::_duplicate (tc);
    tc = tc->unalias();
    if (tc->kind() != CORBA::tk_array)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    CORBA::ULong len = tc->length();
    CORBA::TypeCode_var eltc = tc->content_type();
    for (CORBA::ULong i = 0; i < len; ++i)
	_elements.push_back (_factory()->create_dyn_tc (eltc));
}

DynArray_impl::DynArray_impl (const CORBA::Any &a)
{
    _type = a.type ();
    CORBA::TypeCode_ptr tc = _type->unalias();
    if (tc->kind() != CORBA::tk_array)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());

    CORBA::ULong len = tc->length();
    CORBA::Boolean r = a.array_get_begin ();
    assert (r);
    for (CORBA::ULong i = 0; i < len; ++i) {
	CORBA::Any el;
	r = a.any_get (el);
	assert (r);
	_elements.push_back (_factory()->create_dyn_any (el));
    }
    r = a.array_get_end ();
    assert (r);
}

DynArray_impl::~DynArray_impl ()
{
}

void
DynArray_impl::from_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());

    CORBA::ULong len = tc->unalias()->length();
    CORBA::Boolean r = value.array_get_begin ();
    assert (r);
    for (CORBA::ULong i = 0; i < len; ++i) {
	CORBA::Any el;
	r = value.any_get (el);
	assert (r);
	_elements[i]->from_any (el);
    }
    r = value.array_get_end ();
    assert (r);
}

CORBA::Any*
DynArray_impl::to_any ()
{
    CORBA::Any *a = new CORBA::Any;

    a->type (_type);
    CORBA::Boolean r = a->array_put_begin ();
    assert (r);
    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	CORBA::Any_var el = _elements[i]->to_any ();
	r = a->any_put (el);
	assert (r);
    }
    r = a->array_put_end ();
    assert (r);
    return a;
}

CORBA::AnySeq*
DynArray_impl::get_elements ()
{
    CORBA::AnySeq *seq = new CORBA::AnySeq;
    seq->length (_elements.size());

    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	CORBA::Any_var el = _elements[i]->to_any ();
	(*seq)[i] = el;
    }
    return seq;
}

void
DynArray_impl::set_elements (const CORBA::AnySeq& value)
{
    if (value.length() != _elements.size())
	mico_throw (InvalidSeq ());

    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	_elements[i]->from_any (value[i]);
    }
}


/************************** DynValue *******************************/

/*
 * XXX Shared valuetypes end up in separate copies when using
 *     DynValue. This is due to the way CORBA::Any is implemented.
 */

DynValue_impl::DynValue_impl (const CORBA::Any &a)
{
    _type = a.type();
    CORBA::TypeCode_ptr tc = _type->unalias();
    if (tc->kind() != CORBA::tk_value && tc->kind() != CORBA::tk_value_box)
	mico_throw (CORBA::ORB::InconsistentTypeCode ());
    _isvalue = (tc->kind() == CORBA::tk_value);

    CORBA::Long value_id;
    CORBA::Boolean is_ref;

    if (_isvalue) {
	CORBA::Boolean r = a.value_get_begin (value_id, is_ref);
	assert (r);
    } else {
	CORBA::Boolean r = a.valuebox_get_begin (value_id, is_ref);
	assert (r);
    }

    if (is_ref) {
	assert (!value_id);

	// XXX what shall we do with NIL values??
	assert (0);
    }

    for (CORBA::ULong i = 0; i < tc->member_count_inherited(); ++i) {
	CORBA::Any el;
	CORBA::Boolean r = a.any_get (el);
	assert (r);
	_elements.push_back (_factory()->create_dyn_any (el));
    }

    if (_isvalue) {
	CORBA::Boolean r = a.value_get_end (value_id, is_ref);
	assert (r);
    } else {
	CORBA::Boolean r = a.valuebox_get_end (value_id, is_ref);
	assert (r);
    }
}

DynValue_impl::~DynValue_impl ()
{
}

void
DynValue_impl::from_any (const CORBA::Any& a)
{
    CORBA::TypeCode_var tc = a.type();
    if (!_type->equaltype (tc))
	mico_throw (Invalid ());

    CORBA::Long value_id;
    CORBA::Boolean is_ref;

    if (_isvalue) {
	CORBA::Boolean r = a.value_get_begin (value_id, is_ref);
	assert (r);
    } else {
	CORBA::Boolean r = a.valuebox_get_begin (value_id, is_ref);
	assert (r);
    }

    if (is_ref) {
	assert (!value_id);

	// XXX what shall we do with NIL values??
	assert (0);
    }

    CORBA::TypeCode_ptr tc2 = tc->unalias();
    for (CORBA::ULong i = 0; i < tc2->member_count_inherited(); ++i) {
	CORBA::Any el;
	CORBA::Boolean r = a.any_get (el);
	assert (r);
	_elements[i]->from_any (el);
    }

    if (_isvalue) {
	CORBA::Boolean r = a.value_get_end (value_id, is_ref);
	assert (r);
    } else {
	CORBA::Boolean r = a.valuebox_get_end (value_id, is_ref);
	assert (r);
    }
}

CORBA::Any*
DynValue_impl::to_any ()
{
    CORBA::Any *a = new CORBA::Any;

    a->type (_type);

    CORBA::Long value_id;

    if (_isvalue) {
	CORBA::Boolean r = a->value_put_begin (value_id);
	assert (r);
    } else {
	CORBA::Boolean r = a->valuebox_put_begin (value_id);
	assert (r);
    }

    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	CORBA::Any_var el = _elements[i]->to_any ();
	CORBA::Boolean r = a->any_put (el);
	assert (r);
    }
    if (_isvalue) {
	CORBA::Boolean r = a->value_put_end (value_id);
	assert (r);
    } else {
	CORBA::Boolean r = a->valuebox_put_end (value_id);
	assert (r);
    }
    return a;
}

CORBA::Boolean
DynValue_impl::next ()
{
    // XXX skip nonpublic members...
    return DynAny_impl::next();
}

CORBA::Boolean
DynValue_impl::seek (CORBA::Long index)
{
    // XXX skip nonpublic members...
    return DynAny_impl::seek (index);
}

void
DynValue_impl::rewind ()
{
    // XXX skip nonpublic members...
    DynAny_impl::rewind();
}

CORBA::FieldName
DynValue_impl::current_member_name ()
{
    return CORBA::string_dup (_type->unalias()->member_name_inherited(_index));
}

CORBA::TCKind
DynValue_impl::current_member_kind ()
{
    // XXX unlias ???
    CORBA::TypeCode_var tc = _type->unalias()->member_type_inherited (_index);
    return tc->kind();
}

CORBA::Visibility
DynValue_impl::current_member_visibility ()
{
    return _type->unalias()->member_visibility_inherited (_index);
}

CORBA::NameValueAccessSeq*
DynValue_impl::get_members ()
{
    // XXX cannot read nonpublic members...
    CORBA::NameValueAccessSeq *seq = new CORBA::NameValueAccessSeq;
    seq->length (_elements.size());
    CORBA::TypeCode_ptr tc = _type->unalias ();
    for (CORBA::ULong i = 0; i < _elements.size(); ++i) {
	(*seq)[i].id = tc->member_name_inherited (i);
	(*seq)[i].access = tc->member_visibility_inherited (i);
	CORBA::Any_var a = _elements[i]->to_any ();
	(*seq)[i].value = a;
    }
    return seq;
}

void
DynValue_impl::set_members (const CORBA::NameValueAccessSeq& value)
{
    CORBA::TypeCode_ptr tc = _type->unalias ();
    if (value.length() != tc->member_count())
	mico_throw (InvalidSeq ());

    // XXX cannot write nonpublic members...
    for (CORBA::ULong i = 0; i < value.length(); ++i) {
	if (strcmp (tc->member_name_inherited(i), value[i].id))
	    mico_throw (InvalidSeq ());
	if (tc->member_visibility_inherited(i) != value[i].access)
	    mico_throw (InvalidSeq ());
	// XXX throws Invalid() instead of InvalidSeq() ...
	_elements[i]->from_any (value[i].value);
    }
}


/************************* ORB ***************************/


CORBA::DynAny_ptr
CORBA::ORB::create_dyn_any (const CORBA::Any& value)
{
    CORBA::TypeCode_var tc = value.type();
    switch (tc->unalias()->kind()) {
    case CORBA::tk_void:
    case CORBA::tk_null:
    case CORBA::tk_char:
    case CORBA::tk_wchar:
    case CORBA::tk_boolean:
    case CORBA::tk_octet:
    case CORBA::tk_short:
    case CORBA::tk_ushort:
    case CORBA::tk_long:
    case CORBA::tk_longlong:
    case CORBA::tk_ulong:
    case CORBA::tk_ulonglong:
    case CORBA::tk_float:
    case CORBA::tk_double:
    case CORBA::tk_longdouble:
    case CORBA::tk_any:
    case CORBA::tk_TypeCode:
    case CORBA::tk_objref:
    case CORBA::tk_string:
    case CORBA::tk_wstring:
	return CORBA::DynBasic::_duplicate (new DynBasic_impl (value));
    case CORBA::tk_enum:
	return CORBA::DynEnum::_duplicate (new DynEnum_impl (value));
    case CORBA::tk_struct:
    case CORBA::tk_except:
	return CORBA::DynStruct::_duplicate (new DynStruct_impl (value));
    case CORBA::tk_union:
	return CORBA::DynUnion::_duplicate (new DynUnion_impl (value));
    case CORBA::tk_sequence:
	return CORBA::DynSequence::_duplicate (
	    new DynSequence_impl (value));
    case CORBA::tk_array:
	return CORBA::DynArray::_duplicate (new DynArray_impl (value));
    case CORBA::tk_fixed:
	return CORBA::DynFixed::_duplicate (new DynFixed_impl (value));
    case CORBA::tk_value:
    case CORBA::tk_value_box:
	return CORBA::DynValue::_duplicate (new DynValue_impl (value));
    default:
	assert (0);
    }
    return 0;
}

CORBA::DynAny_ptr
CORBA::ORB::create_dyn_tc (CORBA::TypeCode_ptr type)
{
    switch (type->unalias()->kind()) {
    case CORBA::tk_void:
    case CORBA::tk_null:
    case CORBA::tk_char:
    case CORBA::tk_wchar:
    case CORBA::tk_boolean:
    case CORBA::tk_octet:
    case CORBA::tk_short:
    case CORBA::tk_ushort:
    case CORBA::tk_long:
    case CORBA::tk_longlong:
    case CORBA::tk_ulong:
    case CORBA::tk_ulonglong:
    case CORBA::tk_float:
    case CORBA::tk_double:
    case CORBA::tk_longdouble:
    case CORBA::tk_any:
    case CORBA::tk_TypeCode:
    case CORBA::tk_objref:
    case CORBA::tk_string:
    case CORBA::tk_wstring:
	return CORBA::DynBasic::_duplicate (new DynBasic_impl (type));
    case CORBA::tk_enum:
	return CORBA::DynEnum::_duplicate (new DynEnum_impl (type));
    case CORBA::tk_struct:
    case CORBA::tk_except:
	return CORBA::DynStruct::_duplicate (new DynStruct_impl (type));
    case CORBA::tk_union:
	return CORBA::DynUnion::_duplicate (new DynUnion_impl (type));
    case CORBA::tk_sequence:
	return CORBA::DynSequence::_duplicate (
	    new DynSequence_impl (type));
    case CORBA::tk_array:
	return CORBA::DynArray::_duplicate (new DynArray_impl (type));
    case CORBA::tk_fixed:
	return CORBA::DynFixed::_duplicate (new DynFixed_impl (type));
    default:
	assert (0);
    }
    return 0;
}

CORBA::DynAny_ptr
CORBA::ORB::create_basic_dyn_any (CORBA::TypeCode_ptr type)
{
    return CORBA::DynAny::_duplicate (new DynBasic_impl (type));
}

CORBA::DynStruct_ptr
CORBA::ORB::create_dyn_struct (CORBA::TypeCode_ptr type)
{
    return CORBA::DynStruct::_duplicate (new DynStruct_impl (type));
}

CORBA::DynSequence_ptr
CORBA::ORB::create_dyn_sequence (CORBA::TypeCode_ptr type)
{
    return CORBA::DynSequence::_duplicate (new DynSequence_impl (type));
}

CORBA::DynArray_ptr
CORBA::ORB::create_dyn_array (CORBA::TypeCode_ptr type)
{
    return CORBA::DynArray::_duplicate (new DynArray_impl (type));
}

CORBA::DynUnion_ptr
CORBA::ORB::create_dyn_union (CORBA::TypeCode_ptr type)
{
    return CORBA::DynUnion::_duplicate (new DynUnion_impl (type));
}

CORBA::DynEnum_ptr
CORBA::ORB::create_dyn_enum (CORBA::TypeCode_ptr type)
{
    return CORBA::DynEnum::_duplicate (new DynEnum_impl (type));
}

CORBA::DynFixed_ptr
CORBA::ORB::create_dyn_fixed (CORBA::TypeCode_ptr type)
{
    return CORBA::DynFixed::_duplicate (new DynFixed_impl (type));
}
